@@ -2,7 +2,7 @@ from .agent import Agent
 import geopandas as gpd
 import pandas as pd
 import numpy
-from shapely import geometry
+from shapely import geometry, distance
 import fiona
 import copy
 
@@ -123,8 +123,8 @@ class Model:
                 self.space.at[loc_index, "agents"].append(agent)
                 self.space.at[loc_index, "n_agents"] += 1
                 agent.location_index = loc_index
-            self._manage_class_variables(agent)
-        except ValueError:
+            self._manage_class_variables_add_agent(agent)
+        except:
             print("Error: check that the object in the agents list is of type or a child of Agent")
 
 
@@ -148,7 +148,7 @@ class Model:
             try:
                 for a in agents:
                     self.add_agent(a, loc_index)
-            except ValueError:
+            except:
                 print("Error: check that all objects in the agents list are of type or children of Agent")
 
         else:
@@ -183,7 +183,7 @@ class Model:
                             list_temp.append(a)
                 return list_temp
         
-        except ValueError:
+        except:
             print("Error: Invalid ID")
 
     def agents_with_props(self, condition: str or function):
@@ -227,7 +227,7 @@ class Model:
         """
         return(self.space.at[loc_index, 'agents'])
 
-    def move_agents(self, agents: list or Agent, new_loc_index: int):
+    def move_agent(self, agent: Agent, new_loc_index: int):
         """
         Move an agent to a new location
 
@@ -242,20 +242,64 @@ class Model:
         -------
         N/A
         """
-        if type(agents) is list:
-            for a in agents:
-                if a.location_index is not None:
-                    self.space.at[a.location_index, "agents"].remove(a)
-                a.location_index = new_loc_index
-                self.space.at[new_loc_index, "agents"].append(a)
-                self.space.at[new_loc_index, "n_agents"] += 1
-        else:
-            if agents.location_index is not None:
-                self.space.at[agents.location_index, "agents"].remove(agents)
-            agents.location_index = new_loc_index
-            self.space.at[new_loc_index, "agents"].append(agents)
+        try:
+            if agent.location_index is not None:
+                self.space.at[agent.location_index, "agents"].remove(agent)
+                self.space.at[agent.location_index, "n_agents"] -= 1
+                self._manage_class_variables_remove_agent(agent)
+            
+            agent.location_index = new_loc_index
+            self.space.at[new_loc_index, "agents"].append(agent)
             self.space.at[new_loc_index, "n_agents"] += 1
+            self._manage_class_variables_add_agent(agent)
+        except:
+            print("Error: type of agent variable is not Agent")
 
+    def move_agents(self, agents: list or Agent, new_loc_index: int):
+        """
+        Move an agent or a group of agents to a new location
+
+        Parameters
+        ----------
+        agent: Agent object
+            the agent to move to a new location
+        new_loc_index: int
+            the index of the polygon in the sapce to which agents will move
+        
+        Returns
+        -------
+        N/A
+        """
+        if type(agents) is list:
+            for agent in agents:
+                self.move_agent(agent, new_loc_index)
+        else:
+            agent = agents
+            self.move_agent(agent, new_loc_index)
+
+    def remove_agent(self, agent: Agent):
+        """
+        Removes agents from the model
+
+        Parameters
+        ----------
+        agents: list of Agent objects or an Agent object
+            the agents to remove from the model
+        
+        Returns
+        -------
+        N/A
+        """
+        #try:
+        if agent.location_index is not None:
+            self.space.at[agent.location_index, 'agents'].remove(agent)
+            self.space.at[agent.location_index, "n_agents"] -= 1
+        self.agents.remove(agent)
+        self._manage_class_variables_remove_agent(agent)
+        
+        #except:
+        #    print('Error abp.remove_agent(): agent variable is not of type Agent')
+    
     def remove_agents(self, agents: list or Agent):
         """
         Removes agents from the model
@@ -269,18 +313,16 @@ class Model:
         -------
         N/A
         """
-        if type(agents) is list:
-            for a in agents:
-                self.space.at[a.location_index, 'agents'].remove(a)
-                self.space.at[a.location_index, "n_agents"] -= 1
-                self.agents.remove(a)
+        try:
+            if type(agents) is list:
+                for agent in agents:
+                    self.remove_agent(agent)
+            
+            else:
+                agent = agents
+                self.remove_agent(agent)
         
-        elif type(agents) is Agent:
-            self.space.at[agents.location_index, 'agents'].remove(agents)
-            self.space.at[agents.location_index, "n_agents"] -= 1
-            self.agents.remove(agents)
-        
-        else:
+        except:
             print('Error: agents are not of type Agent or a list of objects of type Agent')
 
     def save_space(self, file_directory: str):
@@ -329,7 +371,7 @@ class Model:
             temp_index_a = self.space[self.space['i'] == i]
             temp_index_b = temp_index_a[temp_index_a['j'] == j]
             return temp_index_b.index[0]
-        except ValueError:
+        except:
             print("Error: invalid i or j")
     
     def indices_in_radius(self, centre: Agent or int or geometry.polygon.Polygon, radius: int or float, outline_only: bool=False, return_patches=False):
@@ -413,8 +455,46 @@ class Model:
             the polygons within the radius in Model.space
         """
         self.indices_in_radius(centre, radius, outline_only, return_patches)
+    
+    def distance(self, a:Agent or int or geometry.polygon.Polygon, b:Agent or int or geometry.polygon.Polygon):
+        """
+        finds the distance between patches and agents
 
-    def _manage_class_variables(self, agents: list or Agent):
+        Parameters
+        ----------
+        a: Agent, index or shapely polygon
+            the first object
+        
+        b: Agent, index or shapely polygon
+            the second object
+        
+        Returns
+        -------
+        float
+            distance (units depends on shapefile)
+        """
+        if type(a) == int:
+            patch_a = self.space.at[a, "geometry"]
+            
+        elif type(a) == geometry.polygon.Polygon:
+            patch_a = a
+        else:
+            patch_a = self.space.at[a.location_index, "geometry"]
+        
+        if type(b) == int:
+            patch_b = self.space.at[b, "geometry"]
+            
+        elif type(b) == geometry.polygon.Polygon:
+            patch_b = b
+        else:
+            patch_b = self.space.at[b.location_index, "geometry"]
+
+        centre_a = patch_a.centroid
+        centre_b = patch_b.centroid
+        return distance(centre_a, centre_b)
+
+
+    def _manage_class_variables_add_agent(self, agents: list or Agent):
         """
         Check if the class of the agent is in the model, create space and Model variables and count the agents by class.
         Forces all created variables to a lowercase.
@@ -448,12 +528,13 @@ class Model:
                     # if the class has already been added
                     elif agent.my_class in self.__classes:
                         name_p_l = f"{agent.my_class}s".lower()
-                        self.__dict__[name_p_l].append(agent)
+                        if agent not in self.__dict__[name_p_l]:
+                            self.__dict__[name_p_l].append(agent)
                         if agent.location_index is not None:
                             self.space.at[agent.location_index, name_p_l].append(agent)
                             self.space.at[agent.location_index, f"n_{name_p_l}"] += 1
                         return True
-            except ValueError:
+            except:
                 print("Error: check that all objects in the agents list are of type or children of Agents")
         else:
             try:
@@ -474,10 +555,43 @@ class Model:
                 # if the class has already been added
                 elif agent.my_class in self.__classes:
                     name_p_l = f"{agent.my_class}s".lower()
-                    self.__dict__[name_p_l].append(agent)
+                    if agent not in self.__dict__[name_p_l]:
+                        self.__dict__[name_p_l].append(agent)
                     if agent.location_index is not None:
                         self.space.at[agent.location_index, name_p_l].append(agent)
                         self.space.at[agent.location_index, f"n_{name_p_l}"] += 1
                     return True
-            except ValueError:
+            except:
                 print("Error: check that the agents input variable is either an Agent or a list of Agents")
+
+    def _manage_class_variables_remove_agent(self, agents: list or Agent):
+        """
+        Manage the class space and Model variables when removing an agent.
+
+        Parameters
+        ----------
+        agent: Agent
+            the agent to remove from the model
+
+        Returns
+        -------
+        N/A
+        """
+        try:
+            if type(agents) == list:
+                for agent in agents:
+                    name_p_l = f"{agent.my_class}s".lower()
+                    self.__dict__[name_p_l].remove(agent)
+                    if agent.location_index is not None:
+                        self.space.at[agent.location_index, name_p_l].remove(agent)
+                        self.space.at[agent.location_index, f"n_{name_p_l}"] -= 1
+            else:
+                agent = agents
+                name_p_l = f"{agent.my_class}s".lower()
+                self.__dict__[name_p_l].remove(agent)
+                if agent.location_index is not None:
+                    self.space.at[agent.location_index, name_p_l].remove(agent)
+                    self.space.at[agent.location_index, f"n_{name_p_l}"] -= 1
+        except:
+            print("Error: check that the agents input variable is either an Agent or a list of Agents")
+
